//Update handler for Legend. Written by Barghest (Todd Allen), toddmallen@gmail.com.

/*******************************************************************************
 * Legend is free software licensed under the GNU General Public License (GPL),*
 * available in the attached COPYING and/or license.txt file in the docs 	   *
 * directory, or at www.fsf.org.																   *
 * This program is provided with NO WARRANTY of any kind and is used or run    *
 * entirely at the user's own risk.											   *
 ******************************************************************************/

#include "legend.h"
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>

extern UPDATE::Update_Handler u_handler;
using UTILS::convert;

static void Sigsegv_Handler(int unused);
static void Sigpipe_Handler(int unused);
static void Sigchld_Handler(int unused);


using UTILS::file_exists;
//template<class type_out, class type_in>
//extern type_out convert(const type_in in);

namespace UPDATE
{
	
Update_Handler::Update_Handler()
{
	shut_down = false;
#ifdef DEBUG
	log_console("Initializing update handler.");
#endif
	log_handler = new Log_Handler;
	register_sighandlers();
}

Update_Handler::~Update_Handler()
{
#ifdef DEBUG
	log_console("Terminating update handler.");
#endif
	if(log_handler && log_handler->get_logging_status())
	{
		delete log_handler;
		log_handler = NULL;
	}
	shut_down = true;
}

//Register the signal handlers.
void Update_Handler::register_sighandlers(void)
{
	signal(SIGPIPE, Sigpipe_Handler);
	signal(SIGCHLD, Sigchld_Handler);
	signal(SIGSEGV, Sigsegv_Handler);
}



Log_Handler::Log_Handler()
{
	if(log_name.empty())
	{
		set_logfile();
	}
	log_silent("Log handler initialized.");
}

Log_Handler::~Log_Handler()
{
	log_silent("Terminating log handler.");
}

void Log_Handler::log_standard(string tolog, int recv_level)
{
	//TODO: Make this check levels through whatever module may be installed.
	//For now won't send to output.
	using ERROR::E_Handler;
	
	fstream log_out;
	try
	{
		log_out.open(log_name.c_str(), fstream::out | fstream::app);
		if(!log_out.is_open())
		{
			throw ERROR::E_Handler("Unable to open logfile!");
		}
		log_out << tolog << endl;
		string lstring;
		lstring = "LOG: ";
		lstring += tolog;
#ifdef DEBUG
		log_console(lstring);
#endif
		log_out.close();
	}
	catch(E_Handler e)
	{
		log_console("ERROR: Log_Handler::log_standard: unable to open logfile! Starting new log...");
		set_logfile();
	}
}

void Log_Handler::log_silent(string tolog)
{
	using ERROR::E_Handler;
	fstream log_out;
	try
	{
		log_out.open(log_name.c_str(), fstream::out | fstream::app);
		if(!log_out.is_open())
		{
			throw ERROR::E_Handler("Unable to open logfile!");
		}
		log_out << tolog << endl;
#ifdef DEBUG
		string lstring;
		lstring = "LOG_SILENT: ";
		lstring += tolog;
		log_console(tolog);
#endif
		log_out.close();
	}
	catch(E_Handler e)
	{
		log_console("ERROR: Log_Handler::log_silent: unable to open logfile! Starting new log...");
		log_console(log_name);
		set_logfile();
	}
}

void Log_Handler::log_error(string tolog, int level, int severity)
{
	fstream log_out;
	try
	{
		log_out.open(log_name.c_str(), fstream::out | fstream::app);
		if(!log_out.is_open())
		{
			throw ERROR::E_Handler("Unable to open logfile!");
		}
		string lstring;
		switch(severity)
		{
			case ERROR_MINOR:
			lstring = "NOTICE: ";
			lstring += tolog;
			break;
			case ERROR_MAJOR:
			lstring = "WARNING: ";
			lstring += tolog;
			break;
			case ERROR_CRITICAL:
			lstring = "ERROR: ";
			lstring += tolog;
			break;
			case ERROR_FATAL:
			lstring = "FATAL ERROR: ";
			lstring += tolog;
			break;
			default:
			lstring = "UNDEFINED ERROR: ";
			lstring += tolog;
			break;
		}
		log_out << lstring << endl;
		log_console(lstring);
		log_out.close();
	}
	catch(ERROR::E_Handler e)
	{
		log_console(e.message);
		set_logfile();
	}
}

void Log_Handler::log_error_silent(string tolog, int severity)
{
	using ERROR::E_Handler;
	fstream log_out;
	try
	{
		log_out.open(log_name.c_str(), fstream::out | fstream::app);
		if(!log_out.is_open())
		{
			throw ERROR::E_Handler("Unable to open logfile!");
		}
		string lstring;
		switch(severity)
		{
			case ERROR_MINOR:
			lstring = "NOTICE: ";
			lstring += tolog;
			break;
			case ERROR_MAJOR:
			lstring = "WARNING: ";
			lstring += tolog;
			break;
			case ERROR_CRITICAL:
			lstring = "ERROR: ";
			lstring += tolog;
			break;
			case ERROR_FATAL:
			lstring = "FATAL ERROR: ";
			lstring += tolog;
			break;
			default:
			lstring = "UNDEFINED ERROR: ";
			lstring += tolog;
			break;
		}
		log_out << lstring << endl;
#ifdef DEBUG
		log_console(lstring);
#endif
	}
	catch(E_Handler e)
	{
		log_console(e.message);
		set_logfile();
	}
}

//TODO: Fix minor bug that seems to increment down (rather than up) the new log.
void Log_Handler::set_logfile(void)
{
	string strbuf, filename;
	register int counter, dircounter;
//	int lognum = MAX_LOG_DUMP;
//	struct stat x;
	bool neednew = false;
	using ERROR::E_Handler;
	
#ifdef DEBUG
	log_console("Starting set_logfile");
#endif
	
	//Will keep logfiles in "batches" of 100 to keep the logs from piling up.
	strbuf = LOGPATH;
	try
	{
		//See if the log directory exists. If not create it.
		if(!file_exists(strbuf))
		{
			if(mkdir(strbuf.c_str(), 0770) == -1)
			{
				throw E_Handler("Unable to create log directory!");
			}
			if(chmod(strbuf.c_str(), (S_IRUSR | S_IWUSR | S_IXUSR | S_IXGRP | S_IRGRP)) == -1)
			{
				throw E_Handler("Unable to set log directory permissions!");
			}
			neednew = true;
		}
		strbuf = LOGPATH;
		strbuf += "/1to";
		strbuf += convert <string> (MAX_LOG_DUMP);
		counter = MAX_LOG_DUMP;
		dircounter = 0;
		//Find the correct subdirectory.
#ifdef DEBUG
	log_console("Position 1:");
	log_console("Strbuf is:");
	log_console(strbuf);
#endif
	string outbuf = strbuf;
	if(!file_exists(strbuf))
		neednew = true;
	while(file_exists(strbuf))
	{
		outbuf = strbuf;
		dircounter = counter - MAX_LOG_DUMP;
		strbuf = LOGPATH;
		strbuf += "/";
		strbuf += convert<string>(counter + 1);
		counter += MAX_LOG_DUMP;
		strbuf += "to";
		strbuf += convert<string>(counter);
#ifdef DEBUG
		log_console("After loop:");
		log_console("Strbuf is:");
		log_console(strbuf);
#endif
	}
		filename = outbuf;
		strbuf = outbuf;
#ifdef DEBUG
		log_console("Strbuf now is: ");
		log_console(strbuf);
#endif
//		if(dircounter > MAX_LOG_DUMP)
//			dircounter -= MAX_LOG_DUMP;
		filename += "/";
		filename += convert<string>(dircounter + 1);
		filename += ".log";
#ifdef DEBUG
		log_console("Filename is:");
		log_console(filename);
#endif
		bool found = false;
		if(!file_exists(filename))
		{
			found = true;
		}
		//Create the next logfile, ###.log.
		if(!found)
		{
#ifdef DEBUG
			log_console("Found was false.");
			log_console("Filename before is: ");
			log_console(filename);
#endif
			for(; dircounter <= counter; dircounter++)
			{
				filename = strbuf;
				filename += "/";
				filename += convert<string>(dircounter);
				filename += ".log";
				if(!file_exists(filename))
				{
					found = true;
					break;
				}
			}
			//Previous log directory was full. Create a new one and start the
			//first log.
			if(!found)
			{
				neednew = true;
				strbuf = LOGPATH;
				strbuf += "/";
				strbuf += convert<string>(counter + 1);
				strbuf += "to";
				strbuf += convert<string>(counter + 101);
			}
		}
#ifdef DEBUG
		log_console("Filename after is:");
		log_console(filename);
#endif
		if(neednew)
		{
			if((mkdir(strbuf.c_str(), 0770) == -1))
			{
				throw E_Handler("Unable to create log directory!");
			}
			if((chmod(strbuf.c_str(), 0750)) == -1)
			{
				throw E_Handler("Unable to set permissions on new directory!");
			}
		}
		log_name = filename;
		log_silent("Beginning logging.");
		logging_status = true;
	}
	catch(E_Handler e)
	{
		log_console(e.message);
		logging_status = false;
		return;
	}
}
				

}//end namespace UPDATE

//Signal handlers.
static void Sigpipe_Handler(int unused)
{
	if((!u_handler.get_loghandler()) || (!u_handler.get_loghandler()->get_logging_status()))
	{
		log_console("ERROR: Caught SIGPIPE.");
	}
	else
	{
		u_handler.get_loghandler()->log_error("ERROR: Caught SIGPIPE.", 0, ERROR_MAJOR);
	}
	return;
}

static void Sigchld_Handler(int unused)
{
	waitpid(-1, NULL, WNOHANG);
}

static void Sigsegv_Handler(int unused)
{
	log_console("ERROR: Segmentation fault!");
	abort();
}
